import copy
import functools
import numpy as np
import pandas as pd
import warnings

from scipy.stats import uniform
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import make_scorer
from sklearn.model_selection import RandomizedSearchCV, GridSearchCV, train_test_split
from xgboost import XGBRegressor

from ..eval import _plot_defaults, get_scores, UpliftEval
from ..explore.base import _add_bins, _NWOE, _NIV, _NIV_bootstrap, _plot_NWOE_bins, _plot_NIV_bs

class BaseProxyMethod:
    """Provide common functionalities for all label transformation methods.

    Requires an input function `transform_func` that transforms `treatment` and
    `outcome` into a single `transformed_outcome`. This is typically the TOT
    transformation, but can be whatever you want.

    Also complete a number of tasks that enable use of the proxy method: save
    dataframe and important dataframe column names to class object, calculate
    the transformed outcome, create an `untransform` method that undoes
    `transform`.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing the features, test/control flag, and outcome.
    transform_func : function
        Function that takes two keyword arguments, `treatment` and
        `outcome`, and outputs a transformed outcome.
    untransform_func : function
        Function that inverts `transform_func`.
    col_treatment : string, optional
        Name of the treatment column. Depends on input dataframe.
    col_outcome : string, optional
        Name of the original outcome column. Depends on input dataframe.
    col_transformed_outcome : string, optional
        Name of the new, transformed outcome. Can be whatever you want.
    col_policy : string or float, optional
        Name of the column that indicates treatment policy (probability of
        treatment). If a float is given, the treatment policy is assumed to be
        even across all rows. If not given, it is assumed that application of
        treatment was randomly assigned with the same probability across the
        entire population.
    continuous_outcome : "infer" or Bool, optional
        Flag that indicates whether or not the Outcome column is continuous.
        Inferred by default.
    random_state : int
        Random seed for deterministic behavior.
    test_size : float
        test_size parameter for skleran.metrics.train_test_split.
    stratify : column name or anything that can be passed to parameter of same
    name in train_test_split, optional
        If not None, stratify is used as input into train_test_split.
    sklearn_model : scikit-learn regressor
        Model used for grid searching and fitting.

    """

    def _score(self, y_true, y_pred, method, plot_type, score_name):
        """ scoring function to be passed to make_scorer.

        """
        treatment_true, outcome_true, p = self.untransform(y_true)
        scores = get_scores(treatment_true, outcome_true, y_pred, p, scoring_range=(0,self.scoring_cutoff[method]), plot_type=plot_type)
        return scores[score_name]

    def __init__(self, df, transform_func, untransform_func, col_treatment='Treatment', col_outcome='Outcome', col_transformed_outcome='TransformedOutcome', col_policy=None, continuous_outcome='infer', random_state=2701, test_size=0.2, stratify=None, scoring_cutoff=1, scoring_method='aqini', sklearn_model=XGBRegressor):

        # ACCOUNT FOR MODULARITY.
        # To allow for modular behavior, different types can be allowed for a
        # number of the above keyword arguments. Below, the snippets of code
        # that account for this modularity are labeled as AFMB(#), which stands
        # for Account for Modular Behavior.

        # Explicitly define some flags that infer the state specified by the
        # modular inputs. I.e. get the `type` and interpret what it means.
        train_test_split_specified = type(df) == tuple
        self.individual_policy_given = type(col_policy) == str

        # AFMB(1): Is train test split specified?
        # If train test split is specified, combine and create a single dataframe.
        if train_test_split_specified:
            df_train = df[0]
            df_test = df[1]
            df = pd.concat([df_train, df_test], keys=[0,1])

        # AFMB(2): Is a policy for each individual? If so, is it a row-level
        # specification?
        # Deal with possible inputs and raise exception if not acceptable.
        if not col_policy:
            treatment = df[col_treatment]
            self.p = len(treatment[treatment==1])/len(treatment)
        elif self.individual_policy_given:
            self.p = df[col_policy]
        elif type(col_policy) == float:  # In the case that it is a numerical value.
            self.p = col_policy
        else:
            raise Exception('col_policy must be a str (column name), float (probability of treatment), or None.')

        # AFMB(3): Is `df[col_outcome]` a continuous variable?
        # By default, this is inferred from the number of distinct values of
        # the Outcome column.
        if continuous_outcome == 'infer':
            num_outcomes = df[col_outcome].nunique()
            continuous_outcome = num_outcomes > 2  # Assign True or False depending on the number of outcomes.

        # AFMB(2) and (3) have an interaction problem in the scenario where
        # both a continuous variable is given for outcome and the policy is
        # specified by row. If this happens, raise exception.
        if continuous_outcome and self.individual_policy_given:
            raise Exception('Cannot give continuous outcome and row-level treatment policy (col_policy).')


        # BEGIN UNMODULAR BEHAVIOR.

        # Save some inputs as class attributes.
        self.random_state = random_state
        self.transform = transform_func
        self.untransform = untransform_func

        # Perform outcome transformation.
        df[col_transformed_outcome] = self.transform(treatment=df[col_treatment], outcome=df[col_outcome], p=self.p)
        # Train test split (or recover the original train test split).
        if train_test_split_specified:
            df_train = df.xs(0)
            df_test = df.xs(1)
        else:
            df_train, df_test = train_test_split(df, test_size=test_size, random_state=self.random_state, stratify=stratify)

        # Save data to class attributes.
        # Full dataframes.
        self.df = df
        self.df_train = df_train
        self.df_test = df_test
        # Column names.
        self.col_treatment = col_treatment
        self.col_outcome = col_outcome
        self.col_transformed_outcome = col_transformed_outcome
        # Transformed outcome.
        self.transformed_y_train = df_train[col_transformed_outcome]
        self.transformed_y_test = df_test[col_transformed_outcome]
        self.transformed_y = df[col_transformed_outcome]
        # Untransformed outcome.
        self.y_train = df_train[col_outcome]
        self.y_test = df_test[col_outcome]
        self.y = df[col_outcome]
        # Features.
        feature_columns = [column for column in df.columns if column not in [col_treatment, col_outcome, col_transformed_outcome, col_policy]]
        self.x_train = df_train[feature_columns]
        self.x_test = df_test[feature_columns]
        self.x = df[feature_columns]
        # Treatment.
        self.tc_train = df_train[col_treatment]
        self.tc_test = df_test[col_treatment]
        self.tc = df[col_treatment]
        # Policy.
        if self.individual_policy_given:
            self.p_train = df_train[col_policy]
            self.p_test = df_test[col_policy]
        else:
            self.p_train = self.p
            self.p_test = self.p
        self.sklearn_model = sklearn_model

        # For backwards compatibility, create plot functions using the plot_type in the function name.
        plot_types = ['qini', 'aqini', 'cgains', 'cuplift', 'uplift', 'balance']
        for plot_type in plot_types:
            new_func = functools.partial(self.plot, plot_type=plot_type)
            setattr(self, 'plot_'+plot_type, new_func)

        # Define scoring cutoff for scoring functions.
        all_scoring_methods = ['qini', 'cgains', 'aqini', 'max_qini', 'max_cgains', 'max_aqini']
        if isinstance(scoring_cutoff, (int, float)):
            self.scoring_cutoff = { method: scoring_cutoff for method in all_scoring_methods }
        elif type(scoring_cutoff) == dict:
            self.scoring_cutoff = scoring_cutoff
        # Define scoring functions.
        for method in all_scoring_methods:
            score_func_name = '_'+method+'_score'
            # Deal with `max` functions differently.
            if method[:3]=='max':
                score_name = method
                plot_type = method[4:]
            else:
                score_name = 'q1_'+method
                plot_type = method
            # Customize `_score` function with `method`.
            new_func = functools.partial(self._score, method=method, plot_type=plot_type, score_name=score_name)
            new_func.__name__ = score_func_name
            setattr(self, score_func_name, new_func)

        # Put scoring functions in form accepted by sklearn functions.
        if type(scoring_method)==str:
            score_func_name = '_'+scoring_method+'_score'
            scoring_arg = make_scorer(getattr(self, score_func_name))
        elif type(scoring_method)==list:
            scoring_arg = {}
            for method in scoring_method:
                if method in all_scoring_methods:
                    score_func_name = '_'+method+'_score'
                    scoring_arg.update({method: make_scorer(getattr(self, score_func_name))})
                else:
                    scoring_arg.update({method: method})  # Allow RandomizedSearchCV built-in scoring methods.

        # Pass scoring function as default arguments in hyperparameter search functions.
        default_params = {
            'verbose': 3,
            'scoring': scoring_arg,
            'refit': False,
        }
        # Default XGBRegressor params.
        if self.sklearn_model == XGBRegressor:
            min_colsamp = np.amax([1/len(self.x_train.columns), 0.3])
            self.randomized_search_params = {
                'estimator': self.sklearn_model(nthread=1),
                'param_distributions': {
                    'n_estimators': range(10,500),
                    'max_depth': list(range(2, 21, 1)),
                    'min_child_weight': list(range(1,500,1)),
                    'gamma': uniform(0, 10),
                    'subsample': uniform(0.3, 0.7),
                    'colsample_bytree': uniform(min_colsamp, 1-min_colsamp),
                },
                'n_iter': 200,
                **default_params
            }
            self.grid_search_params = {
                'estimator': self.sklearn_model(),
                'param_grid': {'min_child_weight': list(range(1,200,1))},
                **default_params
            }
        # Default parameters for other models.
        else:
            self.randomized_search_params = {
                'estimator': self.sklearn_model(),
                **default_params
                }
            self.grid_search_params = {
                'estimator': self.sklearn_model(),
                **default_params
            }



    def NWOE(self, feats_to_use=None, n_bins=10):
        """Net weight of evidence.

        Parameters
        ----------
        feats_to_use : list, optional
            A list of features to use. If no list is specified, all features are used.
        n_bins : int, optional
            Number of bins to use.

        Returns
        -------
        ax : matplotlib.axes._subplots.AxesSubplot
            The matplotlib axes handle.

        """
        feats = feats_to_use if feats_to_use else self.x_train.columns
        df_with_bins = _add_bins(self.df_train, feats, n_bins=n_bins)
        self.NWOE_dict = _NWOE(df_with_bins, feats, col_treatment=self.col_treatment, col_outcome=self.col_outcome)
        ax = _plot_NWOE_bins(self.NWOE_dict, feats)
        return ax

    def NIV(self, feats_to_use=None, n_bins=10, n_iter=3):
        """Net information value, calculated for each feature averaged over `n_iter` bootstrappings of `df`.

        Parameters
        ----------
        feats_to_use : list, optional
            A list of features to use. If no list is specified, all features are used.
        n_bins : int, optional
            Number of bins to use.
        n_iter : int, optional
            Number of iterations.

        Returns
        -------
        ax : matplotlib.axes._subplots.AxesSubplot
            The matplotlib axes handle.

        """
        feats = feats_to_use if feats_to_use else self.x_train.columns
        df_with_bins = _add_bins(self.df_train, feats, n_bins=n_bins)
        # self.NIV_dict = _NIV(df_with_bins, feats, col_treatment=self.col_treatment, col_outcome=self.col_outcome)
        means_dict, low_perc_dict, high_perc_dict = _NIV_bootstrap(self.df_train, feats, n_bins=n_bins, perc=[20,80], n_iter=n_iter, frac=0.5, col_treatment=self.col_treatment, col_outcome=self.col_outcome)
        self.NIV_dict = means_dict
        ax = _plot_NIV_bs(means_dict, low_perc_dict, high_perc_dict, feats)
        return ax

    def randomized_search(self, **kwargs):
        """Randomized search using sklearn.model_selection.RandomizedSearchCV.

        Any parameters typically associated with RandomizedSearchCV (see
        sklearn documentation) can be passed as keyword arguments to this
        function.

        The final dictionary used for the randomized search is saved to
        `self.randomized_search_params`. This is updated with any parameters
        that are passed.

        Examples
        --------
        # Passing kwargs.
        self.randomized_search(param_distributions={'max_depth':[2,3,5,10]}, refit=True)

        """
        self.randomized_search_params.update(kwargs)
        self.rand_search_ = RandomizedSearchCV(**self.randomized_search_params)
        self.rand_search_.fit(self.x_train, self.transformed_y_train)
        return self.rand_search_

    def grid_search(self, **kwargs):
        """Grid search using sklearn.model_selection.GridSearchCV.

        Any parameters typically associated with GridSearchCV (see
        sklearn documentation) can be passed as keyword arguments to this
        function.

        The final dictionary used for the grid search is saved to
        `self.grid_search_params`. This is updated with any parameters that are
        passed.

        Examples
        --------
        # Passing kwargs.
        self.grid_search(param_grid={'max_depth':[2,3,5,10]}, refit=True)

        """
        self.grid_search_params.update(kwargs)
        self.grid_search_ = GridSearchCV(**self.grid_search_params)
        self.grid_search_.fit(self.x_train, self.transformed_y_train)
        return self.grid_search_


    def fit(self, productionize=False, **kwargs):
        """A fit wrapper around any sklearn Regressor.

        Any parameters typically associated with the model can be passed
        as keyword arguments to this function.

        The sklearn model object is saved to `self.model`, or if
        `productionize=True`, `self.model_final`.

        Parameters
        ----------
        productionize: boolean, optional
            If False, fits the model over the train set only. Otherwise, fits
            to all data available.

        """
        if productionize:
            # Just calculate the full new model.
            self.model_final = self.sklearn_model(**kwargs)
            self.model_final.fit(self.x, self.transformed_y)
        else:
            self.model = self.sklearn_model(**kwargs)
            self.model.fit(self.x_train, self.transformed_y_train)

            # Calculate evaluation metrics on test set.
            self.transformed_y_test_pred = self.model.predict(self.x_test)
            self.test_results_ = UpliftEval(self.tc_test, self.y_test, self.transformed_y_test_pred, p=self.p_test)

            # Calculate evaluation metrics for train set.
            self.transformed_y_train_pred = self.model.predict(self.x_train)
            self.train_results_ = UpliftEval(self.tc_train, self.y_train, self.transformed_y_train_pred, p=self.p_train)

    def noise_fit(self, iterations=10, n_bins=10, **kwargs):
        """Shuffle predictions to get a sense of the range of possible curves you
        might expect from fitting to noise.

        Parameters
        ----------
        iterations : int, optional
            Number of times to shuffle the data and retrain.
        n_bins : int, optional
            Number of bins to use when calculating the qini curves.

        """
        noise_fits = []
        shuffled_predictions = copy.deepcopy(self.transformed_y_test_pred)
        for i in range(iterations):
            np.random.shuffle(shuffled_predictions)
            upev = UpliftEval(self.tc_test, self.y_test, shuffled_predictions, n_bins=n_bins, p=self.p_test)
            noise_fits.append(upev)
        self.noise_fits = noise_fits

    def shuffle_fit(self, iterations=10, n_bins=20, params=None, transform_train=None, clear=False, plot_type='cgains', stratify=None, starting_seed=0, **kwargs):
        """Try the train-test split `iterations` times, and fit a model using `params`.

        Parameters
        ----------
        iterations : int
            Number of shuffle-fit sequences to run.
        n_bins : int
            Number of bins for the resulting curve to have.
        params : dict
            Dictionary of parameters to pass to each fit. If not given, will
            default to `self.rand_search_.best_params_`.
        transform_train : func, optional
            A function that will be applied to the training data only. Extended
            functionality that may be useful if the distribution of the
            y-variable is heavy-tailed, and a transformation would produce a
            better model, but you still want to evaluate on the untransformed
            data.
        clear : boolean, optional
            Data for the shuffle fits is saved in `self.shuffle_fit_`. If clear
            is True, this data is rewritten with each shuffle_fit iteration.
        plot_type : string, optional
            Type of plot to show. Can be `aqini`, `qini`, `cgains`.
        stratify : anything that can be passed to parameter of same name in train_test_split, optional
            If not None, stratify is used as input into train_test_split.
        starting_seed : the random seed used for the first iteration of train_test_split. All subsequent iterations increment from this value.

        """
        bootstrap_fits = []

        if not params:
            if hasattr(self, 'rand_search_'):
                if hasattr(self.rand_search_, 'best_params_'):
                    params = self.rand_search_.best_params_
                else:
                    print('ERROR: You need to pass parameters to shuffle fit.')
            else:
                print('ERROR: You need to pass parameters to shuffle fit.')

        ups = {}
        for seed in range(starting_seed, starting_seed+iterations):
            df_train, df_test = train_test_split(self.df, test_size=0.2, random_state=seed, stratify=stratify)
            if transform_train:
                df_train[self.col_outcome] = transform_train(df_train[self.col_outcome])
            uptmp = self.__class__((df_train, df_test), col_treatment=self.col_treatment, col_outcome=self.col_outcome, random_state=seed)
            uptmp.fit(**params, **kwargs)
            uptmp.test_results_.calc(plot_type=plot_type, n_bins=n_bins)
            ups[seed] = uptmp
            print('Seed', seed, 'finished.')

        # Save bootstrap fits to class.
        if (hasattr(self, 'shuffle_fit_')) and (not clear):
            self.shuffle_fit_.update(ups)
        else:
            self.shuffle_fit_ = ups
        return ups

    def plot(self, plot_type='cgains', ax=None, n_bins=None, show_noise_fits=False, noise_lines_kwargs={}, noise_band_kwargs={}, show_shuffle_fits=False, shuffle_lines_kwargs={}, shuffle_band_kwargs={}, shuffle_avg_line_kwargs={}, *args, **kwargs):
        """ Function to plot all curves.

        args and kwargs are passed to the default
        plot function, inherited from the UpliftEval class.

        Parameters
        ----------
        plot_type : string, optional
            Either 'qini', 'aqini', 'uplift', 'cuplift', or 'balance'.
            'aqini' refers to an adjusted qini plot, 'cuplift' gives a
            cumulative uplift plot. 'balance' gives the test-control balance
            for each of the bins. All others are self-explanatory.
        ax : matplotlib.Axes
            Pass axes to allow for overlaying on top of existing plots.
        n_bins : int, optional
            Number of bins to use for the main plot. This has no bearing on the
            shuffle or shuffle plots, which have to be calculated through
            their respective methods.
        show_noise_fits : bool, optional
            Toggle the display of fits to random noise.
        noise_lines_kwargs : dict, optional
            Kwargs to be passed to the lines that display the different curves
            for each noise fit iteration.
        noise_band_kwargs : dict, optional
            Kwargs to be passed to the colored band that displays the standard
            deviation of the noise fit iterations.
        show_shuffle_fits : bool, optional
            Toggle the display of fits with different train test split seeds.
        shuffle_lines_kwargs : dict, optional
            Kwargs to be passed to the lines that display the different curves
            for each shuffle fit iteration.
        shuffle_band_kwargs : dict, optional
            Kwargs to be passed to the colored band that displays the standard
            deviation of the shuffleped fit iterations.
        shuffle_avg_line_kwargs : dict, optional
            Kwargs to be passed to the average line that displays the average
            value of the shuffleped fit iterations.

        """

        # Define default parameters and update with the kwarg dictionaries.
        noise_lines_dict = {'alpha': 0.1}
        noise_lines_dict.update(noise_lines_kwargs)
        noise_band_dict = {'color':[0,0,0], 'alpha':0.3}
        noise_band_dict.update(noise_band_kwargs)
        shuffle_lines_dict = {'alpha':0.1}
        shuffle_lines_dict.update(shuffle_lines_kwargs)
        shuffle_band_dict = {'color':[0,0,0.4], 'alpha':0.3}
        shuffle_band_dict.update(shuffle_band_kwargs)
        shuffle_avg_line_dict = {'color': [0,0,0]}
        shuffle_avg_line_dict.update(shuffle_avg_line_kwargs)

        # If no axes are given, generate axes.
        if not ax:
            fig, ax = _plot_defaults()

        # If we are showing shuffle fits, an average line is shown, so don't plot the default.
        if not show_shuffle_fits:
            self_n_bins = getattr(self.test_results_, plot_type+'_n_bins')
            self_plot_func = getattr(self.test_results_, 'plot_'+plot_type)

            # Recalculate the values if n_bins is not the same.
            if n_bins and (n_bins!=self_n_bins):
                self.test_results_.calc(plot_type=plot_type, n_bins=n_bins)

            ax = self_plot_func(ax=ax, *args, **kwargs)

        # For qini-style curves, remove the random selection line in shuffle & noise fits.
        if show_noise_fits or show_shuffle_fits:
            extra_kwargs = {}
            if plot_type in ['qini', 'aqini', 'cgains']:
                extra_kwargs = {'show_random_selection': False}

        # Plot noise fits if flagged.
        if show_noise_fits and (hasattr(self,'noise_fits')):
            noise_y_values = []

            # Remove random selection line for qini and aqini curves.
            for upev in self.noise_fits:
                getattr(upev, 'plot_'+plot_type)(ax=ax, **noise_lines_dict, **extra_kwargs)
                noise_y_values.append(getattr(upev, plot_type+'_y'))
            noise_means = np.array(list(map(np.mean, list(zip(*noise_y_values)))))
            noise_std = np.array(list(map(np.std, list(zip(*noise_y_values)))))
            ax.fill_between(getattr(self.noise_fits[0], plot_type+'_x'), noise_means-noise_std, noise_means+noise_std, **noise_band_dict)

        # Plot shuffled fits if flagged.
        shuffle_y_values = []
        if show_shuffle_fits and hasattr(self, 'shuffle_fit_'):
            for key, up in self.shuffle_fit_.items():
                up.plot(plot_type=plot_type, ax=ax, **shuffle_lines_dict, **extra_kwargs)
                shuffle_y_values.append(getattr(up.test_results_, plot_type+'_y'))
            bs_means = np.array(list(map(np.mean, list(zip(*shuffle_y_values)))))
            bs_std = np.array(list(map(np.std, list(zip(*shuffle_y_values)))))
            x = getattr(self.shuffle_fit_[0].test_results_, plot_type+'_x')
            ax.fill_between(x, bs_means-bs_std, bs_means+bs_std, **shuffle_band_dict)
            ax.plot(x, bs_means, **shuffle_avg_line_dict)
            ax.plot([x[0], x[-1]], [bs_means[0], bs_means[-1]], '--', color=[0.6,0.6,0.6])

        return ax




import pandas as pd
import numpy as np
import matplotlib.pylab as plt

from ..style import _plot_defaults

def _add_bins(df, feats, n_bins=10):
    """Finds n_bins bins of equal size for each feature in dataframe and outputs the result as a dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
        dataframe with features
    feats : list
        list of features you would like to consider for splitting into bins (the ones you want to evaluate NWOE, NIV etc for)
    n_bins = number of even sized (no. of data points) bins to use for each feature (this is chosen based on both t and c datasets)

    Returns
    ----------
    df_new : pandas.DataFrame
         original dataframe with bin intervals for each feature included as new columns (labelled as original column name + '_bin')
    """

    df_new = df.copy()

    for feat in feats:
        # check number of unique values of feature -- if low (close to the number of bins), we need to be careful
        num_unique_elements = len(df[feat].unique())

        # we should be more careful with how we make bins
        # we really want to make this independent of bins
        if num_unique_elements > n_bins*2: # x2 because we need intervals
            bin_intervals = pd.qcut(df[feat],n_bins,duplicates='drop') # !!! make sure there's nothing funny happening with duplicates
            # include bins in new column
            df_new[str(feat)+'_bin'] = bin_intervals
        else:
            df_new[str(feat)+'_bin'] = df_new[feat]

    return df_new

def _conv_dict_to_array(a_dict, feats):
    """Converts a_dict to an array

    """
    return [a_dict(feat) for feat in feats]

def _get_counts(df_new, feats, col_treatment='Treatment', col_outcome='Outcome'):
    """Gets all of the counts across the intervals as a dictionary for each feature

    Parameters
    ----------
    df_new : pandas.DataFrame
        the original dataframe of data df with bins included using the _add_bins function
    feats: list
        list of feats to consider

    Returns
    -------
    counts_dict : dictionary
        a dictionary of counts used in other functions to calculate NWOE etc. with keys = feature names and values = dataframe of counts

    """

    counts_dict = {}
    y = df_new[col_treatment]
    trt = df_new[col_outcome]
    for feat in feats:
        bin_feat = str(feat)+'_bin'
        counts1_t1 = df_new[(y>0)&(trt==1)][[feat,bin_feat]].groupby(bin_feat).count().rename(columns={feat:'counts_y1t1'})
        counts1_t0 = df_new[(y>0)&(trt==0)][[feat,bin_feat]].groupby(bin_feat).count().rename(columns={feat:'counts_y1t0'})
        counts0_t1 = df_new[(y==0)&(trt==1)][[feat,bin_feat]].groupby(bin_feat).count().rename(columns={feat:'counts_y0t1'})
        counts0_t0 = df_new[(y==0)&(trt==0)][[feat,bin_feat]].groupby(bin_feat).count().rename(columns={feat:'counts_y0t0'})
        # creating a dataframe with all of these results
        counts_dict[feat] = pd.concat([counts1_t1,counts1_t0,counts0_t1,counts0_t0],axis=1).fillna(0)+1 # replace any empty slots with zeros (and add 1 to everything)

    return counts_dict

def _WOE(df_new, feats, trt, col_treatment='Treatment', col_outcome='Outcome'):
    """The WOE for our dataset (with trt telling us which treatment to consider)

    Parameters
    ----------
    df_new : pandas.DataFrame
        original dataframe with bin intervals for each feature included as new columns
    feats : list
        features of interest
    trt : int
        0 or 1 indicating treatment.

    Returns
    -------
    WOE
        Weight of evidence for each feature by bin (a dictionary with keys = feature names)
    """

    # get the counts for each y & trt pair
    counts_dict = _get_counts(df_new, feats, col_treatment=col_treatment, col_outcome=col_outcome)

    # create dictionary WOE of each feature
    WOE = {}

    for feat in feats:
        # calculate the WOE for each bin
        if trt==1:
            WOE_indiv = pd.DataFrame({'WOE':np.log(counts_dict[feat]['counts_y1t1']/counts_dict[feat]['counts_y0t1']*sum(counts_dict[feat]['counts_y0t1'])/sum(counts_dict[feat]['counts_y1t1']))})
        elif trt==0:
            WOE_indiv = pd.DataFrame({'WOE':np.log(counts_dict[feat]['counts_y1t0']/counts_dict[feat]['counts_y0t0']*sum(counts_dict[feat]['counts_y0t0'])/sum(counts_dict[feat]['counts_y1t0']))})
        # could also add a general case for no trt value -- add both t0 t1 columns

        WOE[feat] = WOE_indiv

    return WOE

def _NWOE(df_new, feats, col_treatment='Treatment', col_outcome='Outcome'):
    """
    Net Weight of Evidence by feature and bin

    Parameters
    ----------
    df_new : pandas.DataFrame
        original dataframe with bin intervals for each feature included as new columns
    feats : list
        features of interest

    Returns
    -------
    NWOE :
        the Net Weight of Evidence (weighted by density as each interval can
        hold different amounts of data -- this is more meaningful) for each
        feature by interval (a dictionary with keys = feature names)

    """

    # first get the bins -- we want to use the same bins for trt and contr
    # using both sets of data together should help get the intervals more accurately

    # getting the counts again
    counts_dict = _get_counts(df_new, feats, col_treatment=col_treatment, col_outcome=col_outcome)

    # get the WOEs for trt=0 and trt=1
    WOEs_trt1 = _WOE(df_new, feats, trt=1, col_treatment=col_treatment, col_outcome=col_outcome)
    WOEs_trt0 = _WOE(df_new, feats, trt=0, col_treatment=col_treatment, col_outcome=col_outcome)

    # combine into NWOE dictionary
    NWOE = {}
    for feat in feats:
        #NWOE[feat] = WOE_df['WOE_trt1'] - WOE_df['WOE_trt0']
        # include density -- get counts
        counts = counts_dict[feat]['counts_y0t0']+counts_dict[feat]['counts_y0t1']+counts_dict[feat]['counts_y1t0']+counts_dict[feat]['counts_y1t1']

        NWOE[feat] = pd.DataFrame({'NWOE':(WOEs_trt1[feat] - WOEs_trt0[feat])['WOE']*counts/sum(counts)})

    return NWOE

# note that with the modified NWOE we now already include the density -- so should be removed
def _NIV(df_new, feats, col_treatment='Treatment', col_outcome='Outcome'):
    """Net Information Value by feature

    Parameters
    ----------
    - df_new = original dataframe with bin intervals for each feature included as new columns
    - feats = features of interest

    Returns
    -------
    - NIV = Net Information Value for each feature (a dictionary with keys = feature names)

    """
    NWOE_dict = _NWOE(df_new, feats, col_treatment=col_treatment, col_outcome=col_outcome)

    # calculating the normalization for probabilities
    # this requires overall count of number of people with y=0,1 in trt=0,1
    trt = df_new[col_treatment]
    y = df_new[col_outcome]
    ny0t0 = len(df_new[(y==0)&(trt==0)])
    ny0t1 = len(df_new[(y==0)&(trt==1)])
    ny1t0 = len(df_new[(y>0)&(trt==0)])
    ny1t1 = len(df_new[(y>0)&(trt==1)])

    # get overall counts
    counts_dict = _get_counts(df_new, feats, col_treatment=col_treatment, col_outcome=col_outcome)

    NIV = {}
    for feat in feats:
        # get the counts for y=1,y=0 & trt=1,trt=0 for each feature by bin
        # combine into front term
        NIV_weight = (counts_dict[feat]['counts_y1t1']*counts_dict[feat]['counts_y0t0']/(sum(counts_dict[feat]['counts_y1t1'])*sum(counts_dict[feat]['counts_y0t0'])) - counts_dict[feat]['counts_y1t0']*counts_dict[feat]['counts_y0t1']/(sum(counts_dict[feat]['counts_y1t0'])*sum(counts_dict[feat]['counts_y0t1'])))
        # get NWOE & combine in one df
        NIV_feat = 100*(NIV_weight*NWOE_dict[feat]['NWOE']).sum() # already included the density term in NWOE!
        # We don't need to x 100, but it makes the numbers closer to 1 -- just a convention

        NIV[feat] = NIV_feat

    return NIV

def _NIV_bootstrap(df, feats, n_bins=10, perc=[20,80], n_iter=100, frac=0.5, col_treatment='Treatment', col_outcome='Outcome'):
    """
    Calculates the NIV for each using bootstrapped samples with means, lower and upper percentiles
    to be used for determine which features to use.

    Parameters
    ----------
    - df = the training dataframe with labels 'y', treatment label 'trt', features columns names 'feats' without bins included
    - feats = features of interest
    - perc = list with upper and lower percentiles to calculate from the bootstrapped NIV
    - n_iter = number of bootstraps to take
    - frac = percentage of samples to use for each bootstrap

    Returns
    -------
    - means_dict, low_perc_dict, high_perc_dict = dictionary of means, low percentile, high percentile for each feature across the boostrapped samples
    """

    # array of NIV dictionaries (one for each iteration)
    NIV_dict_array = []

    for i in np.arange(n_iter):
        # including bins for each subset of the dataset
        df_sub_bins = _add_bins(df.sample(frac=frac), feats, n_bins=n_bins)
        # finding the NIV for this subset
        NIV_dict = _NIV(df_sub_bins,feats, col_treatment=col_treatment, col_outcome=col_outcome)
        NIV_dict_array.append(NIV_dict)

    # replacing with a dictionary of arrays
    NIV_array_dict = {}
    for feat in feats:
        NIV_array_dict[feat] = []
        for i in np.arange(len(NIV_dict_array)):
            NIV_array_dict[feat].append(NIV_dict_array[i][feat])

    # creating dictionary of means, lower percentile, upper percentile
    means_dict = {}
    low_perc_dict = {}
    high_perc_dict = {}

    for feat in feats:
        bs_array = np.array(NIV_array_dict[feat])
        # replace any infs with 0
        bs_array[bs_array==np.inf]=0
        # mean values
        means_dict[feat] = np.mean(bs_array)
        low_perc_dict[feat] = np.percentile(bs_array,perc[0])
        high_perc_dict[feat] = np.percentile(bs_array,perc[1])

    return means_dict, low_perc_dict, high_perc_dict

def _plot_NWOE_bins(NWOE_dict, feats):
    """
    Plots the NWOE by bin for the subset of features interested in (form of list)

    Parameters
    ----------
    - NWOE_dict = dictionary output of `NWOE` function
    - feats = list of features to plot NWOE for

    Returns
    -------
    - plots of NWOE for each feature by bin
    """

    for feat in feats:
        fig, ax = _plot_defaults()
        feat_df = NWOE_dict[feat].reset_index()
        plt.bar(range(len(feat_df)), feat_df['NWOE'], tick_label=feat_df[str(feat)+'_bin'], color='k', alpha=0.5)
        plt.xticks(rotation='vertical')
        ax.set_title('NWOE by bin for '+str(feat))
        ax.set_xlabel('Bin Interval');
    return ax

def _plot_NIV_bs(means_dict, low_perc_dict, high_perc_dict, feats):
    """
    Plots the NWOE by bin for the subset of features interested in (form of list)

    Parameters
    ----------
    - NWOE_dict = dictionary output of `NWOE` function
    - feats = list of features to plot NWOE for

    Returns
    -------
    - plots of NWOE for each feature by bin
    """

    # find order of features from highest mean value to lowest
    # could also order by low_perc_dict
    feats_sorted = sorted(means_dict, key=means_dict.get, reverse=False)

    # convert to arrays
    means = np.array([means_dict[feat] for feat in feats_sorted])
    low_perc = np.array([low_perc_dict[feat] for feat in feats_sorted])
    high_perc = np.array([high_perc_dict[feat] for feat in feats_sorted])

    ind = np.arange(len(feats))  # the x locations for the feats
    fig, ax = _plot_defaults(figsize=(15,len(feats)))
    ax.barh(ind, means, xerr=[means-low_perc,high_perc-means], align='center', alpha=0.5, ecolor='black', capsize=8)
    ax.set_ylim([-0.75,len(feats)-0.25])
    ax.set_yticks(ind)
    ax.set_yticklabels(feats_sorted, minor=False)
    ax.set_ylabel('Features')
    ax.set_xlabel('NIV')
    return ax



import numpy as np
import unittest

from pylift.methods.derivatives import TransformedOutcome

treatment = np.array([1,1,0,1,0])
outcome = np.array([1,1,1,0,0])
p = np.array([0.75,0.75,0.75,0.75,0.5])
transformed_outcome = TransformedOutcome._transform_func(treatment, outcome, p)
treatment_new, outcome_new, p_new = TransformedOutcome._untransform_func(transformed_outcome)

class TransformedOutcomeTransformation(unittest.TestCase):
    """Test transformations and inverse transformations for data recovery.

    """

    def test_untransform_treatment(self):
        self.assertTrue(np.isclose(treatment, treatment_new).all())

    def test_untransform_outcome(self):
        self.assertTrue(np.isclose(outcome, outcome_new).all())

    def test_untransform_p(self):
        self.assertTrue(np.isclose(p, p_new).all())

if __name__ == '__main__':
    unittest.main()



import copy
import functools
import numpy as np
import matplotlib.pyplot as plt, matplotlib as mpl
from .style import _plot_defaults

EPS = 1e-20

_ = lambda x: x

def _ensure_array(arraylike):
    """ Make sure inputs are np.array.

    A number of functions exploit np.array's element-level operations, so we
    add this snippet of code to ensure an object is a numpy array.

    Parameters
    ----------
    arraylike : array-like
        Array-like object to be processed.

    Returns
    -------
    arraylike : np.array
        Numpy array with elements from `arraylike`.

    """
    if type(arraylike)!=np.ndarray:
        arraylike = np.array(arraylike)
    return arraylike

def _get_counts(treatment, outcome, p):
    """Extract (treatment,outcome) pair counts from treatment and outcome vectors.

    Calculate counts of outcome/treatment combinations. Variables are named
    nt#o#, where the # corresponds to a binary value for test and outcome,
    respectively.

    Parameters
    ----------
    treatment : np.array
        Vector of 1s and 0s indicating exposure to a treatment.
    outcome : np.array
        Vector of 1s and 0s indicating successful outcome.
    p : np.array
        Vector of values 0 < p < 1 indicating probability of treatment.

    Returns
    -------
    Nt1o1 : int
        Number of entries where (treatment, outcome) == (1,1).
    Nt0o1 : int
        Number of entries where (treatment, outcome) == (0,1).
    Nt1o0 : int
        Number of entries where (treatment, outcome) == (1,0).
    Nt0o0 : int
        Number of entries where (treatment, outcome) == (0,0).

    """
    Nt1o1 = 0.5*np.sum(1/p[(treatment == 1) & (outcome > 0)])
    Nt0o1 = 0.5*np.sum(1/(1-p[(treatment == 0) & (outcome > 0)]))
    Nt1o0 = 0.5*np.sum(1/p[(treatment == 1) & (outcome == 0)])
    Nt0o0 = 0.5*np.sum(1/(1-p[(treatment == 0) & (outcome == 0)]))
    return Nt1o1, Nt0o1, Nt1o0, Nt0o0

def _get_tc_counts(Nt1o1, Nt0o1, Nt1o0, Nt0o0):
    """Get treatment and control group sizes from `_get_counts` output.

    Parameters
    ----------
    Nt1o1 : int
        Number of entries where (treatment, outcome) == (1,1).
    Nt0o1 : int
        Number of entries where (treatment, outcome) == (0,1).
    Nt1o0 : int
        Number of entries where (treatment, outcome) == (1,0).
    Nt0o0 : int
        Number of entries where (treatment, outcome) == (0,0).

    Returns
    -------
    Nt1 : int
        Size of treatment group.
    Nt0 : int
        Size of control group.
    N : int
        Size of full group.

    """
    Nt1 = Nt1o0 + Nt1o1
    Nt0 = Nt0o0 + Nt0o1
    N = Nt0 + Nt1
    return Nt1, Nt0, N

def _get_no_sure_thing_counts(Nt1o1, Nt0o1, Nt1o0, Nt0o0, func=sum):
    """Calculate counts of causal behavior under assumption: no "sure things".

    Determining the number of persuadables, sure things, lost causes, and
    sleeping dogs is impossible, but we can do so if we make assumptions and
    follow these assumptions to their logical end. We can then calculate
    maximal qini curves.

    In this case, we assume there are no "sure things" -- in other words, all
    those in the treatment group that have a positive outcome are "persuadables".

    Parameters
    ----------
    Nt1o1 : int
        Number of entries where (treatment, outcome) == (1,1).
    Nt0o1 : int
        Number of entries where (treatment, outcome) == (0,1).
    Nt1o0 : int
        Number of entries where (treatment, outcome) == (1,0).
    Nt0o0 : int
        Number of entries where (treatment, outcome) == (0,0).
    func : function
        Function to process tuples before output is returned. `sum` is typical,
        but an identity function can be passed if you want to keep the tuples.

    Returns
    -------
    persuadables : func(tuple)
        Number of persuadables in control and treatment, respectively, under the given assumption.
    dogs : func(tuple)
        Number of sleeping dogs, in control and treatment, respectively, under the given assumption.
    sure_things : func(tuple)
        Number of sure things, in control and treatment, respectively, under the given assumption.
    lost_causes : func(tuple)
        Number of lost causes, in control and treatment, respectively, under the given assumption.

    """
    Nt1, Nt0, N = _get_tc_counts(Nt1o1, Nt0o1, Nt1o0, Nt0o0)
    sure_things = (0, 0)
    persuadables = (Nt1o1*Nt0/Nt1, Nt1o1)  # Add the persuadables in control.
    dogs = (Nt0o1, Nt0o1*Nt1/Nt0)
    lost_causes = (Nt0-dogs[0]-sure_things[0]-persuadables[0], \
        Nt1-dogs[1]-sure_things[1]-persuadables[1])
    return func(persuadables), func(dogs), func(sure_things), func(lost_causes)

def _get_no_sleeping_dog_counts(Nt1o1, Nt0o1, Nt1o0, Nt0o0, func=sum):
    """Calculate counts of causal behavior under assumption: no "sleeping dogs".

    Determining the number of persuadables, sure things, lost causes, and
    sleeping dogs is impossible, but we can do so if we make assumptions and
    follow these assumptions to their logical end. We can then calculate
    maximal qini curves.

    In this case, we assume there are no "sleeping dogs" -- in other words, all
    those in the control group that have a positive outcome are "sure things".

    Parameters
    ----------
    Nt1o1 : int
        Number of entries where (treatment, outcome) == (1,1).
    Nt0o1 : int
        Number of entries where (treatment, outcome) == (0,1).
    Nt1o0 : int
        Number of entries where (treatment, outcome) == (1,0).
    Nt0o0 : int
        Number of entries where (treatment, outcome) == (0,0).
    func : function
        Function to process tuples before output is returned. `sum` is typical,
        but an identity function can be passed if you want to keep the tuples.

    Returns
    -------
    persuadables : func(tuple)
        Number of persuadables in control and treatment, respectively, under the given assumption.
    dogs : func(tuple)
        Number of sleeping dogs, in control and treatment, respectively, under the given assumption.
    sure_things : func(tuple)
        Number of sure things, in control and treatment, respectively, under the given assumption.
    lost_causes : func(tuple)
        Number of lost causes, in control and treatment, respectively, under the given assumption.

    """
    Nt1, Nt0, N = _get_tc_counts(Nt1o1, Nt0o1, Nt1o0, Nt0o0)
    dogs = (0,0)
    sure_things = (Nt0o1, Nt0o1*Nt1/Nt0)
    lost_causes = (Nt1o0*Nt0/Nt1, Nt1o0)
    persuadables = (Nt0-dogs[0]-sure_things[0]-lost_causes[0], \
        Nt1-dogs[1]-sure_things[1]-lost_causes[1])
    return func(persuadables), func(dogs), func(sure_things), func(lost_causes)

def _get_overfit_counts(Nt1o1, Nt0o1, Nt1o0, Nt0o0, func=sum):
    """Calculate counts of causal behavior under assumption: all treated are
    persuadables, and these are the only persuadables.

    Determining the number of persuadables, sure things, lost causes, and
    sleeping dogs is impossible, but we can do so if we make assumptions and
    follow these assumptions to their logical end. We can then calculate
    maximal qini curves.

    In this case, we assume that all those with a positive outcome in the
    treatment group are persuadables, and that these are the only persuadables.
    In almost all circumstances, this combination of counts are impossible, but
    they do provide the points for the absolute maximal qini curve possible.

    Parameters
    ----------
    Nt1o1 : int
        Number of entries where (treatment, outcome) == (1,1).
    Nt0o1 : int
        Number of entries where (treatment, outcome) == (0,1).
    Nt1o0 : int
        Number of entries where (treatment, outcome) == (1,0).
    Nt0o0 : int
        Number of entries where (treatment, outcome) == (0,0).
    func : function
        Function to process tuples before output is returned. `sum` is typical,
        but an identity function can be passed if you want to keep the tuples.

    Returns
    -------
    persuadables : func(tuple)
        Number of persuadables in control and treatment, respectively, under the given assumption.
    dogs : func(tuple)
        Number of sleeping dogs, in control and treatment, respectively, under the given assumption.
    sure_things : func(tuple)
        Number of sure things, in control and treatment, respectively, under the given assumption.
    lost_causes : func(tuple)
        Number of lost causes, in control and treatment, respectively, under the given assumption.

    """
    persuadables = (0, Nt1o1)
    dogs = (Nt0o1, 0)
    sure_things = (0,0)
    lost_causes = (Nt0o0, Nt1o0)
    return func(persuadables), func(dogs), func(sure_things), func(lost_causes)

def _maximal_qini_curve(func, Nt1o1, Nt0o1, Nt1o0, Nt0o0):
    """Use counts of (treatment,outcome) pairs to determine the optimal Qini
    curve.

    Calculates the maximal Qini-style curves using the number of entries in the
    different (treatment,outcome) pairs and the _get_* function, `func`.

    Parameters
    ----------
    func: function
        Function that returns persuadables, etc. counts.
    Nt1o1 : int
        Number of entries where (treatment, outcome) == (1,1).
    Nt0o1 : int
        Number of entries where (treatment, outcome) == (0,1).
    Nt1o0 : int
        Number of entries where (treatment, outcome) == (1,0).
    Nt0o0 : int
        Number of entries where (treatment, outcome) == (0,0).

    Returns
    -------
    x : list
        x-values of the maximal Qini curve.
    y : list
        y-values of the maximal Qini curve.

    """
    Nt1, Nt0, N = _get_tc_counts(Nt1o1, Nt0o1, Nt1o0, Nt0o0)
    persuadables, dogs, sure_things, lost_causes = func(Nt1o1, Nt0o1, Nt1o0, Nt0o0)
    # For the overfit case, this is simply the entire treated group.
    if func == _get_overfit_counts:
        slope = 2
    else:
        slope = 1
    x = [0, persuadables/N, 1-dogs/N, 1]
    # Deal with edge case where number of persuadables is greater than sleeping
    # dogs (common if this is not a treatment/control experiment, but an
    # experiment between two treatments).
    if x[1]>x[2]:
        new_val = (x[1]+x[2])/2
        x[1] = new_val
        x[2] = new_val
    y = [0, x[1]*slope, x[1]*slope, (Nt1o1/Nt1-Nt0o1/Nt0)]

    return x, y

def _maximal_uplift_curve(func, Nt1o1, Nt0o1, Nt1o0, Nt0o0):
    """Use counts of (treatment,outcome) pairs to determine the optimal uplift
    curve.

    Calculates the maximal uplift curves using the number of entries in the
    different (treatment,outcome) pairs and the _get_* function, `func`.

    Parameters
    ----------
    func: function
        Function that returns persuadables, etc. counts.
    Nt1o1 : int
        Number of entries where (treatment, outcome) == (1,1).
    Nt0o1 : int
        Number of entries where (treatment, outcome) == (0,1).
    Nt1o0 : int
        Number of entries where (treatment, outcome) == (1,0).
    Nt0o0 : int
        Number of entries where (treatment, outcome) == (0,0).

    Returns
    -------
    x : list
        x-values of the maximal uplift curve.
    y : list
        y-values of the maximal uplift curve.

    """
    Nt1, Nt0, N = _get_tc_counts(Nt1o1, Nt0o1, Nt1o0, Nt0o0)
    persuadables, dogs, sure_things, lost_causes = func(Nt1o1, Nt0o1, Nt1o0, Nt0o0)
    y = [1, 1, 0, 0, -1, -1]
    x = [0, persuadables/N, persuadables/N, 1-dogs/N, 1-dogs/N, 1]
    return x, y

def _maximal_cuplift_curve(func, Nt1o1, Nt0o1, Nt1o0, Nt0o0):
    """Use counts of (treatment,outcome) pairs to determine the optimal
    cumulative uplift curve.

    Calculates the maximal cumulative uplift curves using the number of entries
    in the different (treatment,outcome) pairs and the _get_* function, `func`.

    Parameters
    ----------
    func: function
        Function that returns persuadables, etc. counts.
    Nt1o1 : int
        Number of entries where (treatment, outcome) == (1,1).
    Nt0o1 : int
        Number of entries where (treatment, outcome) == (0,1).
    Nt1o0 : int
        Number of entries where (treatment, outcome) == (1,0).
    Nt0o0 : int
        Number of entries where (treatment, outcome) == (0,0).

    Returns
    -------
    x : list
        x-values of the maximal cumulative uplift curve.
    y : list
        y-values of the maximal cumulative uplift curve.

    """
    Nt1, Nt0, N = _get_tc_counts(Nt1o1, Nt0o1, Nt1o0, Nt0o0)
    persuadables, dogs, sure_things, lost_causes = func(Nt1o1, Nt0o1, Nt1o0, Nt0o0, func=lambda x: x)
    kink_y = (persuadables[1] + sure_things[1])/(persuadables[1] + sure_things[1] + lost_causes[1]) \
        - (sure_things[0])/(persuadables[0] + sure_things[0] + lost_causes[0])
    y = [1, 1, kink_y, (Nt1o1/Nt1-Nt0o1/Nt0)]
    x = [0, sum(persuadables)/N, 1-sum(dogs)/N, 1]
    return x, y

def get_scores(treatment, outcome, prediction, p, scoring_range=(0,1), plot_type='all'):
    """Calculate AUC scoring metrics.

    Parameters
    ----------
    treatment : array-like
    outcome : array-like
    prediction : array-like
    p : array-like
        Treatment policy (probability of treatment for each row).
    scoring_range : 2-tuple
        Fractional range over which frost score is calculated. First element
        must be less than second, and both must be less than 1.

    Returns
    -------
    scores : dict
        A dictionary containing the following values. Each is also appended
        with `_cgains` and `_aqini` for the corresponding values for the
        cumulative gains curve and adjusted qini curve, respectively.

        q1: Traditional Q score normalized by the theoretical
        maximal qini. Note the theoretical max here goes up with a slope of 2.

        q2: Traditional Q score normalized by the practical maximal qini. This
        curve increases with a slope of 1.

        Q: Area between qini curve and random selection line. This is named
        after the notation in Radcliffe & Surry 2011, but note that they
        normalize their curves differently.

        Q_max: Maximal possible qini score, which is used for normalization
        of qini to get frost score. Only obtainable by overfitting.

        Q_practical_max: Practical maximal qini score, if you are not
        overfitting. This assumes that all (outcome, treatment) = (1,1) were
        persuadables, but that there are also an equal number of persuadables
        in the control group. This is the best possible scenario, but likely
        assumes too few "sure things".

        overall_lift: The lift expected from random application of treatment.

    """
    treatment = _ensure_array(treatment)
    outcome = _ensure_array(outcome)
    prediction = _ensure_array(prediction)
    p = _ensure_array(p)

    Nt1o1, Nt0o1, Nt1o0, Nt0o0 = _get_counts(treatment, outcome, p)
    Nt1, Nt0, N = _get_tc_counts(Nt1o1, Nt0o1, Nt1o0, Nt0o0)

    def riemann(x, y):
        avgy = [(a+b)/2 for (a,b) in zip(y[:-1], y[1:])]
        dx = [b-a for (a,b) in zip(x[:-1], x[1:])]
        return sum([a*b for (a,b) in zip(dx, avgy)])

    qini_riemann = riemann(*_maximal_qini_curve(_get_overfit_counts, Nt1o1, Nt0o1, Nt1o0, Nt0o0))
    practical_qini_riemann = riemann(*_maximal_qini_curve(_get_no_sure_thing_counts, Nt1o1, Nt0o1, Nt1o0, Nt0o0))

    overall_lift = (Nt1o1/Nt1-Nt0o1/Nt0)
    qini_max = qini_riemann - 0.5*overall_lift
    practical_qini_max = practical_qini_riemann - 0.5*overall_lift

    # The predicted Qini curve.
    # First we need to reorder the y values and y_pred based on this reordering
    # We calculate TOT roughly here so we have a way of distinguishing those that (ordered, treated) and those that (ordered, untreated).
    y = (2*treatment - 1)*outcome

    def sortbyprediction(vec):
        list2 = list(zip(prediction,vec))
        # Sort by prediction.
        list2.sort(key=lambda tup: tup[0], reverse=True) # included the tup[0] because otherwise we run into problems when there are only a few predicted values -- it orders by index i instead -- not what we want!
        # Extract `y`, sorted by prediction.
        _, vec_ordered = zip(*list2)
        return vec_ordered

    y_ordered = sortbyprediction(y)
    tr_ordered = sortbyprediction(treatment)
    p_ordered = sortbyprediction(p)

    def auc(method='qini'):
        # Calculate the area.
        uplift_last = 0
        nt1o1 = 0
        nt0o1 = 0
        nt1 = EPS
        nt0 = EPS
        pred_riemann = 0
        uplifts = []
        for i in range(round(scoring_range[0]*len(treatment)), round(scoring_range[1]*len(treatment))):
            if y_ordered[i] > 0:
                nt1o1 += 0.5*(1/p_ordered[i])
            elif y_ordered[i] < 0:
                nt0o1 += 0.5*(1/(1-p_ordered[i]))

            if tr_ordered[i] == 1:
                nt1 += 0.5*(1/p_ordered[i])
            else:
                nt0 += 0.5*(1/(1-p_ordered[i]))

            if method=='qini':
                uplift_next = nt1o1/Nt1-nt0o1/Nt0
            elif method=='cgains':
                uplift_next = (nt1o1/nt1-nt0o1/nt0)*(nt1+nt0)/N
            elif method=='aqini':
                uplift_next = nt1o1/Nt1-nt0o1*nt1/(nt0*Nt1 + EPS)

            uplifts.append(uplift_next)
            # each point corresponds to an x delta of 1/N
            pred_riemann += 1/2*(uplift_next+uplift_last)/N
            uplift_last = uplift_next

        AUC = pred_riemann - 0.5*overall_lift*(scoring_range[1]**2 - scoring_range[0]**2)
        maxgain = np.amax(uplifts)
        return AUC, maxgain

    # Dictionary to store all scores.
    scores = {}
    # Raw max scores.
    scores['Q_max'] = qini_max
    scores['overall_lift'] = overall_lift
    scores['Q_practical_max'] = practical_qini_max
    if (plot_type=='qini') or (plot_type=='all'):
        # Qini curve scores.
        scores['Q_qini'], scores['max_qini'] = auc(method='qini')
        scores['q1_qini'] = scores['Q_qini']/scores['Q_max']
        scores['q2_qini'] = scores['Q_qini']/scores['Q_practical_max']
    if (plot_type=='cgains') or (plot_type=='all'):
        # Scores for cumulative gains curve.
        scores['Q_cgains'], scores['max_cgains'] = auc(method='cgains')
        scores['q1_cgains'] = scores['Q_cgains']/scores['Q_max']
        scores['q2_cgains'] = scores['Q_cgains']/scores['Q_practical_max']
    if (plot_type=='aqini') or (plot_type=='all'):
        # Scores for adjusted qini curve.
        scores['Q_aqini'], scores['max_aqini'] = auc(method='aqini')
        scores['q1_aqini'] = scores['Q_aqini']/scores['Q_max']
        scores['q2_aqini'] = scores['Q_aqini']/scores['Q_practical_max']

    return scores

class UpliftEval:
    """Calculate qini and uplift curves given some input data.

    Requires three input vectors: `treatment`, `outcome`, `prediction`. Generates
    based qini and uplift curves where populations are ranked by `prediction`, and
    uplift is calculated over `treatment` and `outcome`.

    This class can be used independently of the methods in this package, i.e.
    if you want to evaluate the performance of an externally generated model.

    NOTE: the maximal Qini curves do not currently work with continuous outcomes.

    Parameters
    ----------
    treatment : array-like
        Array of 1s and 0s indicating whether a treatment was served.
    outcome : array-like
        Arrays of nonzero values and zeros indicating whether a response occurred.
    prediction : array-like
        Predicted value used to rank.
    p : float, None, or array-like
        The treatment policy, P(treatment==1). Can be a float if uniform across
        all individuals; or an array if individual-dependent.

    """

    def __init__(self, treatment, outcome, prediction, p="infer", n_bins=20):

        # Cast input arrays as np.array.
        self.treatment = _ensure_array(treatment)
        self.outcome = _ensure_array(outcome)
        self.prediction = _ensure_array(prediction)

        # Deal with `p`, in case float or None.
        if type(p) == str:
            if p == "infer":
                self.p = np.ones(self.prediction.shape)*len(self.treatment[self.treatment==1])/len(self.treatment)
        elif type(p) == float:
            self.p = np.ones(self.prediction.shape)*p
        else:  # If array.
            self.p = _ensure_array(p)

        plot_types = ['qini', 'aqini', 'cgains', 'cuplift', 'uplift', 'balance']

        # Counts.
        Nt1o1, Nt0o1, Nt1o0, Nt0o0 = _get_counts(self.treatment, self.outcome, self.p)
        self.N_treat = Nt1o0 + Nt1o1
        self.N_contr = Nt0o0 + Nt0o1
        self.N = self.N_treat + self.N_contr

        # Calculate maximal curves.
        count_functions = [_get_overfit_counts, _get_no_sure_thing_counts, _get_no_sleeping_dog_counts]
        count_names = ['max', 'pmax', 'nosdmax']
        curve_functions = [_maximal_qini_curve, _maximal_uplift_curve, _maximal_cuplift_curve]
        curve_names = ['qini', 'uplift', 'cuplift']
        for count_function, count_name in zip(count_functions, count_names):
            for curve_function, curve_name in zip(curve_functions, curve_names):
                x, y = curve_function(count_function, Nt1o1, Nt0o1, Nt1o0, Nt0o0)
                setattr(self, '{}_{}_x'.format(curve_name, count_name), x)
                setattr(self, '{}_{}_y'.format(curve_name, count_name), y)

        # Calculate curve (qini, uplift, cumulative uplift).
        for plot_type in plot_types:
            self.calc(plot_type=plot_type, n_bins=n_bins)

        # Calculate Q, q1, q2 scores and other metrics.
        scores = get_scores(self.treatment, self.outcome, self.prediction, self.p)
        for key in scores:
            setattr(self, key, scores[key])

        # For backwards compatibility, create a bunch of plotting functions using the plot_type in the function name.
        for plot_type in plot_types:
            new_func = functools.partial(self.plot, plot_type=plot_type)
            setattr(self, 'plot_'+plot_type, new_func)

    def calc(self, plot_type, n_bins=20):
        """Calculate the different curve types.

        Parameters
        ----------
        plot_type: string
            Type of curve to calculate. Options: qini, aqini, cgains, cuplift, balance, uplift.
        n_bins : int, optional
            Number of bins to use.

        Returns
        -------
        percentile: list
            The percentile of the population, calculated from the bins.
        qini_y: list
            The qini value for each of the percentile points.
        """
        setattr(self, plot_type+'_n_bins', n_bins)

        # Create bins.
        bin_range = np.linspace(0, len(self.treatment), n_bins+1).astype(int)

        # Define whether the curve uses all data up to the percentile, or the data within that percentile.
        def noncumulative_subset_func(i):
            return np.isin(list(range(len(self.treatment))), prob_index[bin_range[i]:bin_range[i+1]])
        def cumulative_subset_func(i):
            return np.isin(list(range(len(self.treatment))), prob_index[:bin_range[i+1]])

        subsetting_functions = {
            'qini': cumulative_subset_func,
            'aqini': cumulative_subset_func,
            'cgains': cumulative_subset_func,
            'cuplift': cumulative_subset_func,
            'balance': noncumulative_subset_func,
            'uplift': noncumulative_subset_func,
        }

        # Define the function that is calculated within the above bins.
        y_calculating_functions = {
            'qini': lambda nt1o1, nt0o1, nt1, nt0: nt1o1/self.N_treat - nt0o1/self.N_contr,
            'aqini': lambda nt1o1, nt0o1, nt1, nt0: nt1o1/self.N_treat - nt0o1*nt1/(nt0*self.N_treat + EPS),
            'cgains': lambda nt1o1, nt0o1, nt1, nt0: (nt1o1/(nt1+EPS)- nt0o1/(nt0+EPS))*(nt1+nt0)/self.N,
            'cuplift': lambda nt1o1, nt0o1, nt1, nt0: nt1o1/(nt1+EPS) - nt0o1/(nt0+EPS),
            'uplift': lambda nt1o1, nt0o1, nt1, nt0: nt1o1/(nt1+EPS) - nt0o1/(nt0+EPS),
            'balance': lambda nt1o1, nt0o1, nt1, nt0: nt1/(nt0+nt1+EPS)
        }

        # Initialize output lists.
        x = []
        y = []

        # Sort `self.prediction`, descending, then get the indices in the test set that
        # these correspond to.
        prob_index = np.flip(np.argsort(self.prediction), 0)

        # Calculate qini curve points for each bin.
        for i in range(n_bins):
            current_subset = subsetting_functions[plot_type](i)
            # Get the values of outcome in this subset for test and control.
            treated_subset = (self.treatment==1) & current_subset
            resp_treated = self.outcome[treated_subset]
            untreated_subset = (self.treatment==0) & current_subset
            resp_untreated = self.outcome[untreated_subset]
            # Get the policy for each of these as well.
            p_treated = self.p[treated_subset]
            p_untreated = self.p[untreated_subset]

            # Count the number of correct values (i.e. y==1) within each of these
            # sections as a fraction of total ads shown.
            nt1o1 = np.sum(resp_treated*0.5/p_treated)
            nt0o1 = np.sum(resp_untreated*0.5/(1-p_untreated))
            nt1 = np.sum(0.5/p_treated)
            nt0 = np.sum(0.5/(1-p_untreated))
            y.append(y_calculating_functions[plot_type](nt1o1, nt0o1, nt1, nt0))

            x.append(nt1+nt0)

        # For non-cumulative functions, we need to do a cumulative sum of the x
        # values, because the sums in the loop only captured the counts within
        # the non-cumulative bins.
        if plot_type in ['balance', 'uplift']:
            x = np.cumsum(x)

        # Rescale x so it's between 0 and 1.
        percentile = x/np.amax(x)

        if plot_type not in ['balance', 'uplift', 'cuplift']:
            percentile = np.insert(percentile, 0, 0)
            y.insert(0,0)

        setattr(self, plot_type+'_x', percentile)
        setattr(self, plot_type+'_y', y)

        return percentile, y

    def plot(self, plot_type='cgains', ax=None, show_theoretical_max=False, show_practical_max=False, show_random_selection=True, show_no_dogs=False, **kwargs):
        """Plots the different kinds of percentage-targeted curves.

        Parameters
        ----------
        plot_type : string, optional
            Either 'qini', 'aqini', 'uplift', 'cuplift', or 'balance'.
            'aqini' refers to an adjusted qini plot, 'cuplift' gives a
            cumulative uplift plot. 'balance' gives the test-control balance
            for each of the bins. All others are self-explanatory.
        ax: matplotlib.axes._subplots.AxesSubplot, optional
            A matplotlib axis referencing where to plot.
        show_theoretical_max: boolean, optional
            Toggle theoretical maximal qini curve, if overfitting to
            treatment/control. Only works for Qini-style curves.
        show_practical_max : boolean, optional
            Toggle theoretical maximal qini curve, if not overfitting to
            treatment/control. Only works for Qini-style curves.
        show_no_dogs : boolean, optional
            Toggle theoretical maximal qini curve, if you believe there are no
            sleeping dogs. Only works for Qini-style curves.
        show_random_selection: boolean, optional
            Toggle straight line indicating a random ordering. Only works for
            Qini-style curves.

        """
        if not ax:
            fig, ax = _plot_defaults()

        titles = {
            'qini': 'Qini curve',
            'aqini': 'Adjusted Qini curve',
            'cgains': 'Cumulative gain chart',
            'cuplift': 'Cumulative uplift curve',
            'uplift': 'Uplift curve',
            'balance': 'Treatment balance curve',
        }

        ylabels = {
            'qini': 'Uplift gain',
            'aqini': 'Uplift gain',
            'cgains': 'Uplift gain',
            'cuplift': 'Cumulative lift',
            'uplift': 'Lift',
            'balance': 'Treatment size / (treatment size + control size)',
        }

        ax.plot(getattr(self, plot_type+'_x'), getattr(self, plot_type+'_y'), '.-', **kwargs)

        if (plot_type=='aqini') or (plot_type=='cgains') or (plot_type=='qini'):
            max_plot_type='qini'
            # Plot random selection line.
            if show_random_selection:
                ax.plot([0,1], [0, getattr(self, plot_type+'_y')[-1]], '--', color=[0.6, 0.6, 0.6], label='Random selection')
        else:
            max_plot_type=plot_type

        if plot_type!='balance':
            # Show theoretical maxes.
            if show_theoretical_max:
                ax.plot(getattr(self, max_plot_type+'_max_x'), \
                    getattr(self, max_plot_type+'_max_y'), 'g--', label='Theoretical max')
            if show_practical_max:
                ax.plot(getattr(self, max_plot_type+'_pmax_x'), \
                    getattr(self, max_plot_type+'_pmax_y'), '--', color=[0.7,0.3,0.3], label='Practical max')
            if show_no_dogs:
                ax.plot(getattr(self, max_plot_type+'_nosdmax_x'), \
                    getattr(self, max_plot_type+'_nosdmax_y'), '--', color=[0.5,0.3,0.7], label='No sleeping dogs')

        ax.set_xlabel('Fraction of data')
        ax.set_ylabel(ylabels[plot_type])
        ax.set_title(titles[plot_type])

        ax.legend(frameon=False)
        return ax




import numpy as np
import unittest

from pylift.eval import _get_counts, _get_tc_counts, _get_no_sure_thing_counts, _get_no_sleeping_dog_counts, _get_overfit_counts, get_scores

treatment = np.array([1,1,0,1,0])
outcome = np.array([1,1,1,0,0])
p = np.array([0.75,0.5,0.75,0.75,0.5])
prediction = np.array([1,0.9,0.8,0.5,0.6])
Nt1o1, Nt0o1, Nt1o0, Nt0o0 = _get_counts(treatment, outcome, p)
scores = get_scores(treatment, outcome, prediction, p)

simple_counts_example = (4/3, 2, 2/3, 1)

treatment_simple = [1,0,1,0]
outcome_simple = [1,0,0,1]
prediction_simple = [1, 0.5, 0.25, 0]
p_simple = [0.5, 0.5, 0.5, 0.5]
scores_simple = get_scores(treatment_simple, outcome_simple, prediction_simple, p_simple)

class CountsAndScores(unittest.TestCase):
    """Test all counting and scoring functions.

    """

    # The following tests were calculated with the package and each step was verified.
    def test_Nt1o1(self):
        self.assertAlmostEqual(Nt1o1, 5/3)

    def test_Nt0o1(self):
        self.assertAlmostEqual(Nt0o1, 2)

    def test_Nt1o0(self):
        self.assertAlmostEqual(Nt1o0, 2/3)

    def test_Nt0o0(self):
        self.assertAlmostEqual(Nt0o0, 1)

    def test_aqini(self):
        self.assertAlmostEqual(scores['Q_aqini'], 0.21279762)

    def test_qini(self):
        self.assertAlmostEqual(scores['Q_qini'], 0.18601190)

    def test_cgains(self):
        self.assertAlmostEqual(scores['Q_cgains'], 0.11737351)

    def test_practical_max(self):
        self.assertAlmostEqual(scores['Q_practical_max'], 0.24943311)

    def test_max(self):
        self.assertAlmostEqual(scores['Q_max'], 0.39527530)

    # The following tests were calculated by hand.
    def test_counts(self):
        self.assertTrue(np.allclose(_get_counts(treatment, outcome, p), (5/3, 2, 2/3, 1)))

    # The following were also calculated by hand, but are redundant to earlier
    # tests. These examples are a nice in that they are a bit simpler, though.
    def test_cgains(self):
        self.assertAlmostEqual(scores_simple['Q_cgains'], 0.28125, msg="Incorrect cumulative gains score.")

    def test_aqini(self):
        self.assertAlmostEqual(scores_simple['Q_aqini'], 0.375, msg="Incorrect cumulative gains score.")

    def test_qini(self):
        self.assertAlmostEqual(scores_simple['Q_qini'], 0.375, msg="Incorrect cumulative gains score.")

if __name__ == '__main__':
    unittest.main()



import numpy as np
import pandas as pd
from scipy.stats import norm

def dgp(N=1000, n_features=3, beta=None, error_std=0.5, tau=3, tau_std=1, discrete_outcome=False, seed=None, feature_effect=0.5):
    """
    dgp(N=1000, n_features=3, beta=[1,-2,3,-0.8], error_std=0.5, tau=3, discrete_outcome=False)

    Generates random data with a ground truth data generating process.

    Draws random values for features from [0, 1), errors from a 0-centered
    distribution with std `error_std`, and creates an outcome y.

    Parameters
    ----------
    N : int, optional
        Number of observations.
    n_features : int, optional
        Number of features.
    beta : np.array, optional
        Array of beta coefficients to multiply by X to get y.
    error_std : float, optional
        Standard deviation (scale) of distribution from which errors are drawn.
    tau : float, optional
        Effect of treatment.
    tau_std : float, optional
        When not None, draws tau from a normal distribution centered around tau
        with standard deviation tau_std rather than just using a constant value
        of tau.
    discrete_outcome : boolean, optional
        If True, outcomes are 0 or 1; otherwise continuous.
    seed : int, optional
        Random seed fed to np.random.seed to allow for deterministic behavior.

    Output
    ------
    df : pd.DataFrame
        A DataFrame containing the generated data.

    """

    np.random.seed(seed=seed)

    if(beta and (n_features != len(beta) -1)):
        raise ValueError('If custom beta supplied, len(beta) must be equal to n_features+1')

    # Effect of features on outcome.
    if beta is None:
        beta = np.random.random(n_features+1)

    # Define features, error, and random treatment assignment.
    X = np.random.random(size=(N, n_features))
    error = np.random.normal(size=(N), loc=0, scale=error_std)
    treatment = np.random.binomial(1, .5, size=(N))

    # Treatment heterogeneity.
    tau_vec = np.random.normal(loc=tau, scale=tau_std, size=N) + np.dot(X, beta[1:])*feature_effect

    # Calculate outcome.
    y = beta[0] + np.dot(X, beta[1:]) + error + treatment*tau_vec

    if discrete_outcome:
        y = y > 0

    names = list(range(n_features))
    names.extend(['Treatment', 'Outcome'])

    df = pd.DataFrame(np.concatenate((X, treatment.reshape(-1,1), y.reshape(-1,1)), axis=1), columns=names)
    return df



def sim_pte(N=1000, n_features=20, beta=None, rho=0, sigma=np.sqrt(2), beta_den=4, discrete_outcome=False, seed=None):
    """
    sim_pte(N=1000, p=20, rho=0, sigma=np.sqrt(2), beta_den=4)

    Numerical simulation for treatment effect heterogeneity estimation as described in Tian et al. (2012)
    Translated from the R uplift package (Leo Guelman <leo.guelman@gmail.com>).

    Parameters
    ----------
    N : int, optional
        Number of observations.
    n_features : int, optional
        Number of features.
    beta : np.array, optional
        Array of beta coefficients to multiply by X to get y.
    rho : covariance matrix between predictors.
    sigma : multiplier of error term.
    beta_den : size of main effects relative to interaction effects.
    discrete_outcome : boolean, optional
        If True, outcomes are 0 or 1; otherwise continuous.
    seed : int, optional
        Random seed fed to np.random.seed to allow for deterministic behavior.

    Output
    ------
    A data frame including the response variable (Y), the treatment (treat=1)
    and control (treat=-1) assignment, the predictor variables (X) and the "true"
    treatment effect score (ts).

    """
    p = n_features

    # Check arguments
    if N < 2:
        raise ValueError("uplift: The number of observations must be greater than 2")

    if p < 4:
        raise ValueError("uplift: The number predictors must be equal or greater than 4")

    if rho < 0 | rho > 1:
        raise ValueError("uplift: rho must be between 0 and 1")

    if sigma < 0:
        raise ValueError("uplift: beta.den must be equal or greater than 0")

    if beta_den <= 0:
        raise ValueError("uplift: beta.den must be greater than 0")

    # Main Effects
    if beta is None:
        beta = np.zeros(p)
        for j in range(p):
            beta[j] = (-1)**j * (j >= 3 & j <= 10) / beta_den

    ### Generate x ~ N~p(0, rho)
    mean = np.zeros(p)
    cov = np.identity(n=p)
    x = np.random.multivariate_normal(mean, cov, N)

    ### Random error from N~1(0,1)
    eps = np.random.normal(loc=0, scale=1, size=N)[:,np.newaxis]

    ### Treatment generated with equal probability at random
    treat = np.random.binomial(n=1, p=0.5, size=N)


    ### Interaction effects
    gamma = np.array([0.5, -0.5, 0.5, -0.5] + [0]*(p-4))

    ### Response variable
    y = np.matmul(x, beta[:,np.newaxis]) + (x * treat[:, np.newaxis]) @ gamma[:, np.newaxis] + sigma * eps

    ### "True" score
    ts = norm.pdf(np.matmul(x, (beta + gamma)[:, np.newaxis]) / sigma) - norm.pdf(np.matmul(x, (beta - gamma)[:, np.newaxis]) / sigma)

    if discrete_outcome:
        y = (y > 0).astype(int)

    ### Returned value
    out = pd.DataFrame({'y':y[:,0], 'treat':treat, 'ts':ts[:,0]})
    dfx = pd.DataFrame(x)
    dfx.columns = ['x' + str(colname) for colname in list(dfx.columns)]

    out = pd.concat([out, dfx], axis=1)

    return out




from .base import BaseProxyMethod
from ..eval import get_scores, UpliftEval
import functools
from scipy.stats import uniform
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import make_scorer
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV, train_test_split
from xgboost import XGBClassifier, XGBRegressor

import numpy as np
import pickle

TOL = 1e-37
EPS = 1e-20

def custom_objective(dtrain, preds):
    # Error.
    err = preds - dtrain
    treatment, outcome, p = TransformedOutcome._untransform_func(dtrain)
    # Correct for policy.
    nt_scaled = np.sum(0.5*treatment/p)
    nc_scaled = np.sum(0.5*(1-treatment)/(1-p))
    p_scaled = nt_scaled/(nt_scaled + nc_scaled)

    w_i = 0.25*(treatment/(p_scaled*p) + (1-treatment)/((1-p_scaled)*(1-p)))
    grad = w_i*err
    hess = np.ones(dtrain.shape)
    return grad, hess

class FlaggedFloat(float):
    """Float subclass that retains a Treatment flag property.
    """
    def save(self, outcome, treatment, p):
        self.outcome = outcome
        self.treatment = treatment
        self.p = p

class TransformedOutcome(BaseProxyMethod):
    """Implement Transformed Outcome [Trees] method.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing the features, test/control flag, and outcome.
    col_treatment : string, optional
        Name of the treatment column. Depends on input dataframe.
    col_outcome : string, optional
        Name of the original outcome column. Depends on input dataframe.
    col_transformed_outcome : string, optional
        Name of the new, transformed outcome. Can be whatever you want.
    col_policy : string or float, optional
        Name of the column that indicates treatment policy (probability of
        treatment). If a float is given, the treatment policy is assumed to be
        even across all rows. If not given, it is assumed that application of
        treatment was randomly assigned with the same probability across the
        entire population.
    continuous_outcome : "infer" or Bool, optional
        Flag that indicates whether or not the Outcome column is continuous.
        Inferred by default.
    random_state : int
        Random seed for deterministic behavior.
    test_size : float
        test_size parameter for skleran.metrics.train_test_split.
    stratify : string or same format as parameter of same name in
    train_test_split, optional
        If not None, stratify is used as input into train_test_split.
    scoring_method : string or list, optional
        Either `qini`, `aqini`, `cgains` or `max_` prepended to any of the
        previous values. Any strings available to the parameter `scoring` in
        `sklearn.model_selection.RandomizedSearchCV` can also be passed.
    scoring_cutoff : float or dict, optional
        The fraction of observations used to score qini for hyperparam
        searches. E.g. if 0.4, the 40% of observations with the highest
        predicted uplift are used to determine the frost score in the
        randomized search scoring function. If a list of scoring_methods is
        passed, a dictionary can also be passed here, where the keys are the
        scoring_method strings and the values are the scoring cutoff for those
        specific methods.
    sklearn_model : sklearn regressor class, optional
        Sklearn model object to for all successive operations (don't pass any
        parameters).
    """

    @staticmethod
    def _transform_func(treatment, outcome, p):
        """Function that executes the Transformed Outcome.

        Parameters
        ----------
        treatment : array-like
            Array of 1s and 0s indicating treatment.
        outcome : array-like
            Array of 1s and 0s indicating outcome.
        p : float or np.array
            Probability of observing a treatment=1 flag, used for the
            transformation.

        Returns
        -------
        y : np.array
            Transformed label.

        """
        treatment = list(treatment)
        outcome = list(outcome)
        if np.size(p) == 1:  # If numerical.
            p = [p for i in range(len(treatment))]
        else:  # If array-like.
            p = list(p)

        # Change nonzero outcomes according to treatment/control split.
        y = [FlaggedFloat(o*((t-pr)/(pr*(1-pr)))) for o, t, pr in zip(outcome, treatment, p)]
        for yi, oi, ti, pi in zip(y, outcome, treatment, p):
            yi.save(oi, ti, pi)
        y = np.array(y, dtype=np.dtype(FlaggedFloat))
        return y

    @staticmethod
    def _untransform_func(ys):
        """Function that recovers original data from Transformed Outcome.

        Parameters
        ----------
        ys : numpy.array
            Transformed label.
        p : float or np.array, optional
            Probability of observing a treatment=1 flag, used to reverse the
            transformation.

        Returns
        -------
        treatment : np.array
            Array of 1s and 0s indicating treatment.
        outcome : np.array
            Array of 1s and 0s indicating outcome.

        """
        treatment = np.array([i.treatment for i in ys])
        p = np.array([i.p for i in ys])
        outcome = np.array([i.outcome for i in ys])

        return treatment, outcome, p

    def __init__(self, df, col_treatment='Treatment', col_outcome='Outcome', col_transformed_outcome='TransformedOutcome', col_policy=None, continuous_outcome='infer', random_state=2701, test_size=0.2, stratify=None, scoring_cutoff=1, sklearn_model=XGBRegressor, scoring_method='cgains'):

        super().__init__(df, transform_func=self._transform_func, untransform_func=self._untransform_func, col_treatment=col_treatment, col_outcome=col_outcome, col_transformed_outcome=col_transformed_outcome, col_policy=col_policy, continuous_outcome=continuous_outcome, random_state=random_state, test_size=test_size, stratify=stratify, scoring_cutoff=scoring_cutoff, scoring_method=scoring_method, sklearn_model=sklearn_model)


class TOLeafXGB(TransformedOutcome):

    def __init__



from setuptools import find_packages, setup

setup(name='pylift',
      version='0.1.2',
      description='Python implementation of uplift modeling.',
      author='Robert Yi, Will Frost',
      author_email='robert@ryi.me',
      url='https://github.com/pylift/pylift',
      install_requires=[
            'numpy',
            'matplotlib',
            'scikit-learn',
            'scipy',
            'seaborn',
            'xgboost'
          ],
      packages=find_packages(),
      zip_safe=False)



import matplotlib.pyplot as plt, matplotlib as mpl

def _plot_defaults(figsize=(15,10), **kwargs):
    """Some aesthetically-pleasing default plot parameters.

    """
    # Change fonts to serif.
    mpl.rcParams['mathtext.fontset'] = 'stix'
    mpl.rcParams['font.family'] = 'STIXGeneral'

    mpl.rcParams['font.size'] = 20
    fig, ax = plt.subplots(figsize=figsize, **kwargs)
    return fig, ax
